/* $Id: access.c,v 1.2 2006/09/16 12:17:17 maxim Exp $
 *
 */

#include "configure.h"

#if SUPPORT_ACCESS

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <unistd.h>


#include "common/misc.h"
#include "common/pairs.h"
#include "common/ip_ranges.h"
#include "common/str.h"

#include "access.h"
#include "httpd.h"
#include "core.h"
#include "config.h"


static unsigned char b64[256];


void
base64initialize()
{
	int i;
	if (b64[0]) return;

	memset(b64, 64, sizeof(b64));
	for (i = 'A'; i <= 'Z'; i++)
		b64[i] = i - 'A';
	for (i = 'a'; i <= 'z'; i++)
		b64[i] = i - 'a' + 26;
	for (i = '0'; i <= '9'; i++)
		b64[i] = i - '0' + 52;
	b64['+'] = 62;
	b64['/'] = 63;
	b64['='] = 0;
}

int
base64decode(unsigned char *encoded, unsigned char *decoded)
{
	int c;
	unsigned char t1, t2, u1, u2, u3;

	while (1) {
		c = *encoded++;
		if (c == 0) {
			*decoded = 0;
			return 0;
		}
		t1 = b64[c];
		if (t1 == 64)
			return -1;
		c = *encoded++;
		if (c == 0)
			return -1;
		t2 = b64[c];
		if (t2 == 64)
			return -1;
		u1 = t1 << 2 | t2 >> 4;
		c = *encoded++;
		if (c == 0)
			return -1;
		t1 = b64[c];
		if (t1 == 64)
			return -1;
		u2 = (t2 & 0xf) << 4 | t1 >> 2;
		c = *encoded++;
		if (c == 0)
			return -1;
		t2 = b64[c];
		if (t2 == 64)
			return -1;
		u3 = (t1 & 0x3) << 6 | t2;
		*decoded++ = u1;
		if (u1 == 0)
			return 0;
		*decoded++ = u2;
		if (u2 == 0)
			return 0;
		*decoded++ = u3;
		if (u3 == 0)
			return 0;
	}
}


char
checkPasswd(Access* access, Request* r)
{
	char user[LONG_STR];
	char* password;
	char* crypted;
	char* a;

	if (!r->authorization) return FAILED;

	if (strncasecmp(r->authorization, "basic ", 6))
		return message(ERROR, "unsupported authorization type %s (%s%s): %s", ip2string(r->ip), r->host?:"-", r->uri, r->authorization);

	a = r->authorization + 6;

	if (strlen(a) >= LONG_STR)
		return message(ERROR, "too long authorization from %s (%s%s): %s", ip2string(r->ip), r->host?:"-", r->uri, strlen(a));

	if (base64decode(a, user))
		return message(ERROR, "bad encoded authorization from %s (%s%s): %s", ip2string(r->ip), r->host?:"-", r->uri, a);

	if (!(password = strchr(user, ':')))
		return message(ERROR, "bad authorization format from %s (%s%s): %s", ip2string(r->ip), r->host?:"-", r->uri, user);
	*password++ = 0;

	refreshPasswd(access);

	if (!(crypted = getPair(access->users, user)))
		return message(ERROR, "no such user from %s (%s%s): %s", ip2string(r->ip), r->host?:"-", r->uri, user);

	if (strcmp(crypted, crypt(password, crypted)))
		return message(ERROR, "wrong password from %s (%s%s): %s", ip2string(r->ip), r->host?:"-", r->uri, user);

	r->user = addStringToBuffer(r, user, password - user);
	return OK;
}


int
checkAccess(Request* r, Access* access, char** redirect)
{
	while (access) {
		switch (access->type) {
		case ANY:
			goto found;

#if SUPPORT_COUNTRY
		case COUNTRY:
			if (r->country < 0) r->country = ip2id(countries, r->ip);
			if (r->country == access->country) goto found;
			break;
#endif

		case NETWORK:
			if ((r->ip & access->mask) == access->network) goto found;
			break;

		case PASSWD:
			if (checkPasswd(access, r)) return 0;
			else {
				char auth[LONG_STR];
				int n = catstr(auth, LONG_STR, "Basic realm=\"", access->realm, "\"", NULL);
				r->auth = addStringToBuffer(r, auth, n + 1);
				return 401;
			}

		case REFERRER: {
			char copy[LONG_STR];
			char* referrer = NULL;
			int rl = 0;

			if (r->referrer && !strncmp(r->referrer, "http://", strlen("http://"))) {
				char* slash;
				slash = strchr(r->referrer + strlen("http://"), '/');
				if (slash) {
					rl = slash - (r->referrer + strlen("http://"));
					if (rl >= LONG_STR) rl = LONG_STR - 1;
					memcpy(copy, r->referrer + strlen("http://"), rl);
					copy[rl++] = 0;
					referrer = unifyHostName(copy);
				}
			}

			switch (access->refType) {
			case REFERRER_EMPTY:
				if (!referrer) goto found;
				break;

			case REFERRER_SELF:
				if (referrer && r->host && !strncmp(referrer, r->host, rl)) goto found;
				break;

			case REFERRER_FOREIGN:
				if (referrer && r->host && strncmp(referrer, r->host, rl)) goto found;
				break;

			case REFERRER_SITE:
				if (referrer && !strncmp(referrer, access->referrer, rl)) goto found;
				break;
			}
			break;
		}

		case AGENT:
			if (r->userAgent && strstr(r->userAgent, access->agent)) goto found;
			break;
		}

		access = access->next;
	}

	return 0;

found:
	if (redirect) *redirect = access->redirect;
	return access->code;
}

int
checkIP(unsigned ip, Access* access)
{
	while (access) {
		if ((ip & access->mask) == access->network) return access->code;
		access = access->next;
	}

	return FALSE;
}
#endif // SUPPORT_ACCESS

