/* $Id: httpd.c,v 1.4 2006/09/17 08:46:46 maxim Exp $
 *
 */

#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/time.h>

#include "configure.h"

#include "common/misc.h"
#include "common/settings.h"
#include "common/signames.h"
#include "common/ip_ranges.h"
#if DEBUGLOG
#include "common/ptrinfo.h"
#endif
#include "config.h"
#include "core.h"
#include "httpd.h"



int doRestart, doReload, doShutdown, doQuit, childExits;
time_t start, configured;
#if SUPPORT_COUNTRY
IPRanges* countries;
IPRanges* classes;
IPRanges* regions;
#endif

void actionOnSignal(int sig);

void
reapChildren()
{
	int status, pid;
	signal(SIGCHLD, actionOnSignal);

	while (TRUE) {
		pid = waitpid(-1, &status, WNOHANG);
		if (pid <= 0) break;
		childExits = YES;

		if (WIFEXITED(status))
			message(MESSAGE, "child process %d exited with status %d", pid, WEXITSTATUS(status));
		else if (WIFSIGNALED(status))
			message(MESSAGE, "child process %d killed by signal %s", pid, signame(WTERMSIG(status)));
	}
	if (pid < 0 && errno != ECHILD)
		message(SYS|ERROR, "cannot waitpid");
}

void
actionOnSignal(int sig)
{
	signal(sig, SIG_IGN);
	getTime();
	message(MESSAGE, "receiving signal %s", signame(sig));

	switch(sig) {
	case SIGIO:
		// bad, very bad and very dirty quick-fix
		// of overflowing rt-signals queue
		doRestart = YES;
		break;

	case SIGHUP:
		doRestart = YES;
		break;
	case SIGWINCH:
		break;
	case SIGQUIT:
		doShutdown = YES;
		break;
	case SIGUSR1:
		doReload = YES;
		break;
	case SIGUSR2:
		Debug = Debug ? OFF : ON;
		message(MESSAGE, "turning debug to %s", Debug ? "ON" : "OFF");
		break;
	case SIGCHLD:
		reapChildren();
		break;
	case SIGPIPE:
		break;
	case SIGSEGV:
		if (currentRequest) logRequest(currentRequest);
#ifdef HAVE_BACKTRACE
		{
			void *array[32];
			unsigned short i;
			int n = backtrace(array, 32);
			char **res = backtrace_symbols(array, n);
			message(MESSAGE, "Backtrace:");
			for (i = 0; i < n; i++)
				message(MESSAGE, "	%s", res[i]);
		}
#endif
		abort();

	case SIGINT:
	case SIGTERM:
	case SIGILL:
	case SIGTRAP:
	case SIGBUS:
	case SIGXCPU:
	case SIGXFSZ:
		if (++doQuit > 10) {
			message(ERROR, "quit signal was not handled %d times, immediately dying", doQuit - 1);
			exit(1);
		}
		break;
	}
	signal(sig, actionOnSignal);
}


int
main(int argc, char** argv)
{
	int i;
	char daemon = YES;
	char wrapper = NO;
	progName = "httpd";
	/*
	 struct rlimit rlim;
	 getrlimit(RLIMIT_CORE, &rlim);
	 if (rlim.rlim_cur < (10<<20)) {
	 rlim.rlim_cur = (10<<20);
	 setrlimit(RLIMIT_CORE, &rlim);
	 }
	 */
	signal(SIGINT,	actionOnSignal);
	signal(SIGQUIT, actionOnSignal);
	signal(SIGTERM, actionOnSignal);
	signal(SIGSEGV, actionOnSignal);
	signal(SIGHUP, actionOnSignal);
	signal(SIGUSR1, actionOnSignal);
	signal(SIGUSR2, actionOnSignal);
	signal(SIGWINCH, actionOnSignal);
	signal(SIGILL,	actionOnSignal);
	signal(SIGTRAP, actionOnSignal);
	signal(SIGBUS,	actionOnSignal);
	signal(SIGCHLD, actionOnSignal);
	signal(SIGURG, actionOnSignal);
	signal(SIGIO, actionOnSignal);
	signal(SIGPIPE, SIG_IGN);
	chdir(ROOT);

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			int j;
			for (j = 0; argv[i][j]; j++) {
				switch (argv[i][j]) {
				case 'f': daemon = NO; break;
				case 'w': wrapper = YES; break;
				case 't':
					if (readConfig(argc - 1, argv + 1))
						return puts("Config OK") & 0;
					puts("Config file contains errors, see logs/current/error.log for details");
					return 2;
				case 'v':
					printf("Server version: %s, built: %s\n"
						   "Configured as: ./configure %s\n",
						   VERSION, built, CONFIGURE);
					return 1;
				case 'h':
					printf("Server version: %s, built: %s\n"
						   "Usage: %s [-options]\n"
						   "  -f	   don't daemonize (work in foreground)\n"
						   "  -w	   use wrapper\n"
						   "  -c NAME  use config file NAME\n"
						   "  -t	   test config file and exit\n"
						   "  -v	   show version and exit\n"
						   "  -h	   show this help page and exit\n",
						   VERSION, built, argv[0]);
					return 1;
				}
			}
		}
	}

	if (wrapper) {
		if (!writePidFile((progName = "httpd-wrapper"), daemon, NO)) return 1;

		while(!doQuit) {
			int pid = fork();
			if (pid < 0) return message(SYS|ERROR, "cannot fork"), 1;
			if (!pid) break;
			message(MESSAGE, "launching httpd [%d]", pid);
			while (!childExits) {
				int s = 0;
				if (doQuit) s = SIGTERM;
				else if (doShutdown) s = SIGQUIT;
				else if (doRestart) s = SIGHUP;
				else if (doReload) s = SIGUSR1;
				doReload = doRestart = NO;
				if (s && kill(pid, s)) message(SYS|ERROR, "cannot send signal %s to %d", signame(s), pid);
				pause();
			}
			childExits = NO;
			if (doShutdown) doQuit = YES;

			message(MESSAGE, "httpd exits");
			if (!doQuit) sleep(1);
		}
		if (doQuit) return 0;

		if (!writePidFile((progName = "httpd"), NO, NO)) return 1;
	} else	{
		if (!writePidFile(progName, daemon, NO)) return 1;
	}

	time(&start);

	doRestart = doReload = doShutdown = doQuit = NO;
	if (work(argc - 1, argv + 1)) {
#if DEBUGLOG
        showPtrs(YES);
#endif
		return message(MESSAGE, "%d is shutting down", myPID);
	}

	message(MESSAGE, "%d is exited due to some init error", myPID);
	return 1;
}


