/* $Id: httpd.h,v 1.6 2006/11/18 09:20:55 maxim Exp $
 *
 */

#ifndef HTTPD_H
#define HTTPD_H

#include <limits.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

#include "common/settings.h"
#include "common/pairs.h"
#include "common/ip_ranges.h"

#include "configure.h"

#include "buffer.h"

#define VERSION "0.7p"

#define TRANSFER_PLAIN	  0
#define TRANSFER_SENDFILE 1
#define TRANSFER_MMAP	  2

#define MODEL_SELECT 0
#define MODEL_POLL	 1
#define MODEL_RTSIG	 2
#define MODEL_KEVENT 3


#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
#define HAS_KEVENT
#endif

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#define HAS_BSD_SENDFILE
#endif

#if defined(__linux__)
#define HAS_LINUX_SENDFILE
#define HAS_LINUX_RTSIG
#endif


//
// method of transferring local files
// plain: uses simple read/write,  most compability, low performance
// mmap: maps files and use write, wide compability, good performance
// sendfile:					   linux, freebsd,	 best performance
//

#ifndef TRANSFER_MODE

#if defined(HAS_LINUX_SENDFILE) || defined(HAS_BSD_SENDFILE)
#define TRANSFER_MODE TRANSFER_SENDFILE
#else
#define TRANSFER_MODE TRANSFER_MMAP
#endif

#endif  /* TRANSFER_MODE */

//
//	event notification model (automatically selected)
//	select:	   most compability, low performance  (not ready yet)
//	poll:	   wide compability, good performance
//	rt-signal: Linux 2.4+,		 best performance
//	kevent:	   FreeBSD 4.1+,	 best performance
//


#ifndef MODEL

#if defined(HAS_LINUX_RTSIG)
#define MODEL MODEL_RTSIG
#elif defined(HAS_KEVENT)
#define MODEL MODEL_KEVENT
#else
#define MODEL MODEL_POLL
#endif

#endif  /* MODEL */


//
// low watermark support
//
#define LOWAT_NONE    0
#define LOWAT_SOCKOPT 1
#define LOWAT_KEVENT  2

#ifndef LOWAT_MODE

#if defined(HAS_KEVENT)
#define LOWAT_MODE LOWAT_KEVENT
#else
#define LOWAT_MODE LOWAT_NONE
#endif

#endif


#define EXPIRES_FUTURE INT_MAX
#define EXPIRES_PAST   INT_MIN



#if SUPPORT_PROXY

// maximum number of unknown headers which transmitted to/from backend
#ifndef MAX_OTHER_HEADERS
#define MAX_OTHER_HEADERS 15
#endif

#endif

// initial and maximum input buffer size for request
#define INITIAL_BUFFER_SIZE 1000
#define MAXIMUM_BUFFER_SIZE 16000


// maximum length of URI (including query string)
#define MAX_URI 8000


// limit URI by HUGE_STR (16k)
#if MAX_URI > HUGE_STR
#error too long MAX_URI
#endif

////////////////////////////////////////////////////////////////////


enum {
	UNKNOWN,
	HEAD,
	GET,
	POST
};

#if SUPPORT_REMAP
enum {
	REMAP_NOTHING,
	REMAP_1000,
	REMAP_1000_MOD
};
#endif

#define ERRORS 2


typedef struct SimpleList_struct SimpleList;
typedef struct Config_struct Config;
typedef struct Listener_struct Listener;
typedef struct Server_struct Server;
typedef struct Alias_struct Alias;
#if SUPPORT_CACHE
typedef struct Cache_struct Cache;
typedef struct CacheEntry_struct CacheEntry;
#endif

#if SUPPORT_ACCESS
typedef struct Access_struct Access;
#endif

#if SUPPORT_PROXY
typedef struct Proxy_struct Proxy;
typedef struct ProxyRequest_struct ProxyRequest;
typedef struct Backend_struct Backend;

#if SUPPORT_FILTER
typedef struct Filter_struct Filter;
typedef struct FilterState_struct FilterState;
#endif

#endif
typedef struct CGIServer_struct CGIServer;
typedef struct Request_struct Request;
typedef struct CGIRequest_struct CGIRequest;
typedef union FD_union FD;

struct Config_struct {
	char* configName;
	char* admin;

	int connections;
#if SUPPORT_PROXY
	int proxyConnections;
#endif
#if SUPPORT_ANTIFLOOD
	int connectionsPerIP;
	int requestsPerIP;
	time_t antifloodPeriod;
	time_t antifloodExpired;
	int antiflood;
#endif
	int keepAliveTimeout;
	int requestTimeout;
	int timeout;
	int expires;
	int keepalive;
	int charset;
	int srcCharset;
	int log;
	int listing;
#if SUPPORT_THROTTLE
	int throttle;
#endif
#if SUPPORT_ELAPSED
	int elapsed;
#endif
	int maxBuffer;
	int maxPostSize;
	int skipWWW;
	int skipPort;
	int filesCheck;
	int filesKeep;
	int sendBuffer;
	int lowWatermark;
	int detachRoot;
	int serverTokens;

	Listener* listeners;
	Pairs* extensions;
	SimpleList* mimes;
	SimpleList* indexes;
	SimpleList* reverse;
#if SUPPORT_PATTERNS
	SimpleList* patterns;
#endif
	Server* servers;
	Server* defaultServer;
	Pairs* serverNames;
#if SUPPORT_ACCESS
	Access* access;
	Access* trusted;
#endif
#if SUPPORT_FILTER
	Filter* filter;
#endif
#if SUPPORT_PROXY
	Backend* backends;
#endif
#if SUPPORT_CACHE
	Pairs* caches;
	Cache* cache;
#endif

	int nOfListeners;
	int active;
};

struct Server_struct {
	char* name;
	char* admin;

	int timeout;
	int expires;
	int keepalive;
	int charset;
	int srcCharset;
	int log;
	int listing;
#if SUPPORT_THROTTLE
	int throttle;
#endif
#if SUPPORT_ELAPSED
	int elapsed;
#endif
#if SUPPORT_ANTIFLOOD
	int antiflood;
#endif

	unsigned hash;
	int nOfAlters;
	char** alter;
	unsigned* alterHash;


	SimpleList* indexes;
	SimpleList* reverse;
	Pairs* extensions;
	Alias* aliases;
#if SUPPORT_CACHE
	Cache* cache;
#endif

#if SUPPORT_ACCESS
	Access* access;
#endif
#if SUPPORT_FILTER
	Filter* filter;
#endif
	int requests;
	int sent;

	int requestsLast;
	int sentLast;

	int requestsToday;
	long long sentToday;

	Server* next;
};


struct Alias_struct {
	int timeout;
	int expires;
	int keepalive;
	int log;
	int listing;
#if SUPPORT_THROTTLE
	int throttle;
#endif
#if SUPPORT_ELAPSED
	int elapsed;
#endif
#if SUPPORT_ANTIFLOOD
	int antiflood;
#endif
	int charset;
	int srcCharset;

	int type;
	int length;
	char* alias;
	char* location;
	char* aux[8];
#if SUPPORT_REMAP
	int remap;
#endif

	char exact;
	char exactLocation;

	int nOfTails;
	char** tail;
	int* tailLength;

	int nOfCtype;
	char** ctype;
	int* ctypeLength;

	SimpleList* indexes;
	SimpleList* reverse;
#if SUPPORT_CACHE
	Cache* cache;
#endif
#if SUPPORT_PROXY
	Proxy* proxy;
#endif
	CGIServer* cgi;
#if SUPPORT_ACCESS
	Access* access;
#endif
#if SUPPORT_FILTER
	Filter* filter;
#endif

	Alias* next;
};

struct Listener_struct {
	unsigned char type;
	struct in_addr addr;
	int port;
	int fd;
	unsigned accepts;
	unsigned acceptsLast;
	unsigned acceptsToday;
	unsigned droppedS;
	unsigned droppedC;
	char* error;
	Listener* next;
};


#if SUPPORT_PROXY

struct Backend_struct {
	char* name;
	struct in_addr addr;
	unsigned short port;
	unsigned short state;
	time_t last;
	unsigned nRequests;
	unsigned maxRequests;
	unsigned requests;
	unsigned requestsLast;
	unsigned requestsToday;
	unsigned lastRequests;

	Backend* next;
};


struct Proxy_struct {
	Backend* backend;
	char* location;

	Proxy* next;
};

struct ProxyRequest_struct {
	unsigned char type;
	unsigned char state;

	char watchRead;
	char watchWrite;

	int fd;
	time_t last;


	/*	request parameters */
	time_t start;
	Request* request;
	Proxy* proxy;
	Buffer* buffer;
	int posted;

	/* reply headers */
	char major;
	char minor;
	char keepalive;
	char acceptRanges;
	int status;
	char* contentType;
	char* contentEnc;
	char* location;
	char* setCookie;
	char* auth;
	int length;
	int rangeFrom;
	int rangeTo;
	int rangeTotal;
	time_t modified;
	time_t expires;
#if MAX_OTHER_HEADERS
	char* outHeaders[MAX_OTHER_HEADERS + 1];
#endif
	int offset;

#if SUPPORT_ELAPSED
	struct timeval exactStart;
#endif

	ProxyRequest* prev;
	ProxyRequest* next;
};

#endif

struct CGI_struct {
};


#if SUPPORT_FILTER
struct Filter_struct {
	char* type;
	int typeLength;
	void (*handler)(Request* request, char flush);
};

struct FilterState_struct {
	char state;
	char quote;
	short pos;
};
#endif
struct SimpleList_struct {
	char* name;
	void* link;
	SimpleList* next;
};

#if SUPPORT_ACCESS
struct Access_struct {
	int code;

	int type;

	char* realm;
	char* passwd;
	char* file;
	Pairs* users;
	time_t mtime;

	int network;
	int mask;

#if SUPPORT_COUNTRY
	short country;
	short region;
#endif

	short refType;
	char* referrer;

	char* agent;

	char* redirect;

	Access* next;
};
#endif

struct Request_struct {
	unsigned char type;
	unsigned char state;
	char watchRead;
	char watchWrite;

	Server* server;
	Alias* alias;
	unsigned ip;
	unsigned realIP;
	int fd;
	int file;
	time_t last;
#if SUPPORT_THROTTLE
	short throttled;
#endif
#if SUPPORT_COUNTRY
	short country;
	short region;
#endif
#if SUPPORT_ACCESS
	char trusted;
#endif

	/*	request parameters */
	time_t start;
	short status;
	unsigned char method;
	char major;
	char minor;
	char keepalive;
	char acceptRanges;
	char wasRequest;
#if SUPPORT_CACHE
	char nocache;
#endif
	int pipelined;

	char* uri;
	char* path;
	char* params;
	char* query;
	Buffer* buffer;
	Buffer* postBuffer;

	/*	request headers */
	char* userAgent;
	char* referrer;
	char* authorization;
	char* user;
	char* auth;
	char* cookie;
	char* host;
	char* inContentType;
	int inContentLength;
	time_t ims;
	time_t iums;
	time_t ifRange;
	unsigned forwardedFor;
#if MAX_OTHER_HEADERS
	char* inHeaders[MAX_OTHER_HEADERS + 1];
#endif

	int rangeTotal;
	int rangeFrom;
	int rangeTo;

	/* reply headers */
	char* contentType;
	char* contentEnc;
	char* location;
	char* setCookie;
	int length;
	int fileSize;
	time_t modified;
	time_t expires;
#if MAX_OTHER_HEADERS
	char* outHeaders[MAX_OTHER_HEADERS + 1];
#endif

	char* body;
	char* alloc;
	char* content;
	int offset;
	int received;
#if SUPPORT_FILTER
	int filtered;
#endif
	int sent;
	int contentSize;

	void* mmap;

#if SUPPORT_PROXY
	ProxyRequest* proxy;
#endif
#if SUPPORT_CACHE
	CacheEntry* cached;
#endif

	CGIRequest* cgi;
#if SUPPORT_FILTER
	FilterState filter;
#endif
	Config* config;

	Request* prev;
	Request* next;
};


union FD_union {
	char* type;
	Listener* listener;
	Request* request;
#if SUPPORT_PROXY
	ProxyRequest* proxy;
#endif
	CGIRequest* cgi;
};

struct CGIRequest_struct {
	unsigned char type;
	unsigned char state;

	int fd;
	time_t last;

	/*	request parameters */
	Request* request;
	Buffer* buffer;
	int posted;

	/* reply headers */
	char major;
	char minor;
	char keepalive;
	char acceptRanges;
	int status;
	char* contentType;
	char* contentEnc;
	char* location;
	char* setCookie;
	char* auth;
	int length;
	int rangeFrom;
	int rangeTo;
	int rangeTotal;
	time_t modified;
	time_t expires;
#if MAX_OTHER_HEADERS
	char* outHeaders[MAX_OTHER_HEADERS + 1];
#endif
};


#if SUPPORT_CACHE
struct Cache_struct {
	char* name;

	size_t maxSize;
	time_t maxAge;
	time_t minAge;
	size_t maxLength;
	int skipHost;
	int skipQuery;

	size_t size;
	int count;
	int orphans;
	int creates;
	int hits;

	CacheEntry** entries;
	CacheEntry* recent;
	CacheEntry* old;
};

struct CacheEntry_struct {
	unsigned hash;
	char* name;
	char* contentType;
	char* contentEnc;
	char* body;
	size_t length;
	time_t modified;
	time_t expires;
	time_t created;
	int links;
	int hits;

	CacheEntry* prev;
	CacheEntry* next;

	CacheEntry* recent;
	CacheEntry* old;
};

#endif



extern int doRestart, doReload, doShutdown, doQuit;
extern int myPID;
extern time_t start, configured;
extern char built[];
extern char past[SHORT_STR];
extern char future[SHORT_STR];
#if SUPPORT_COUNTRY
extern IPRanges* countries;
extern IPRanges* regions;
#endif

#endif
